# Embedding models
from chesspos.models.autoencoder import AutoencoderModel
from chesspos.models.dense_autoencoder import DenseAutoencoder
from chesspos.models.resnet_autoencoder import ResnetAutoencoder
from chesspos.models.cnn_autoencoder import CnnAutoencoder
